/* W4Row.c
 * Game by Christian Barmala
 * Object is to get four chips in line
 * For detailled playing instructions and for an explanation of
 * the Minimax algorithm and the CGI interface cf. the attached
 * documentation */


#include <stdio.h>
#include <ctype.h>
#include <string.h>
/* depending on your compiler you must modify the following lines */
#include <stdlib.h>
/* typedef int void; */
/* #define const */

#define MYADDRESS "<A HREF=\"mailto:christb@informix.com (Christian Barmala)\"> Christian Barmala</A>\n"

/* chip contains 0:empty, 1:yellow, 2:red 3:special purpose */
typedef unsigned char chiptype;
#define CEMT 0
#define CYEL 1
#define CRED 2
#define CSPC 3
#define opponent(x) (CSPC - x)

/* columns reaches from 0 ... 6 and rows from 0 ... 5 */
typedef signed char ytype;
typedef signed char xtype;

/* value of a chain of length x 
 * the values of VNUL, ... VTRE can be fine tuned
 * the value of VFOR should be larger than any possible sum
 * of values VNUL, ... VTRE, because this indicates that
 * one party has won already */
#define VNUL 0
#define VONE 1
#define VTWO 10
#define VTRE 100
#define VFOR 10000


#define MAXWINVALUE (VFOR + 1)
/* to keep CPU time low, the maximum strength is limited */
#define MAXSTRENGTH 10
/* the default strength of the computer */
#define DEFAULTSTRENGTH 5


/* Chips are inserted to the top of one of the 7 columns
 * of the rack and fall down to the lowest unoccupied row */
chiptype rack[7][6];
/* number of moves computed in advance */
char strength = DEFAULTSTRENGTH;
/* when computing the possible moves, the computer should
 * check the middle columns first and left and right ones
 * later */
ytype sequence[7] = {6, 0, 1, 5, 4, 2, 3};
ytype lastmoves[2] = {-1, -1}; /* the last moves of the two parties */
char allmoves[42] = {0}; /* all moves of the game */
/* number of empty fields per column */
xtype spcclm[7] = {6, 6, 6, 6, 6, 6, 6};
signed char spcrack = 42; /* number of empty fields in the rack */
xtype actmove; /* actual depth of move ahead, always < strength */
chiptype winner; /* which player has won already */
chiptype player;


#define inschip(x, chiptoinsert) rack[x][--(spcclm[x])] = chiptoinsert;
#define remchip(x) rack[x][(spcclm[x])++] = CEMT;


/* Print Error */
void error(s)
const char * const s;
{
	printf("Content-type: text/html\n\n");
	printf("<HTML>\n<HEAD>\n<TITLE>W4Row Error</TITLE>\n</HEAD>\n<BODY>\n");
	printf("<H1>W4Row Error</H1>\n%s\n<ADDRESS>%s</ADDRESS>\n", s, MYADDRESS);
	printf("</BODY>\n</HTML>\n");
}


/* macros used in value()
 * watch out: no white space after "\"! */

/* fill shift registers regyel/regred with "1" for
 * every yellow/red chip */
#define shiftchip() \
{ \
	regyel = ((regyel << 1) & 0x0f) | (chip & 1); \
	regred = ((regred << 1) & 0x0f) | ((chip >> 1) & 1); \
}


/* fill both shift registers regyel/regred with "1" meaning this
 * position is occupied and not usable by neither parties */
#define shiftocc() \
{ \
	regyel = ((regyel << 1) & 0x0f) | 1; \
	regred = ((regred << 1) & 0x0f) | 1; \
}


#if (VNUL == 0)
#define EVALCONDRED (!regred && regyel)
#define EVALCONDYEL (!regyel && regred)
#else
#define EVALCONDRED !regred
#define EVALCONDYEL !regyel
#endif


/* evaluate red and yellow pattern if not chip of the oposite
 * color blocks the way */
#define evalpat() \
{ \
	if(EVALCONDRED){ \
		valyel += convpat[regyel]; \
		if(valyel >= VFOR) return VFOR; \
	} \
	if(EVALCONDYEL){ \
		valred += convpat[regred]; \
		if(valred >= VFOR) return -VFOR; \
	} \
}


short int value()
{
	/* define 4 bit shift registers. Initialization not required! */
	register unsigned char regyel, regred;
	char x, y, z; /* row and column and auxiliary counter */
        chiptype chip;
	/* value from yellow or red point of view */
	short int valyel = 0, valred = 0;
	/* convert bitpattern to values */
	static const short int convpat[16] =
		{VNUL, VONE, VONE, VTWO,
		 VONE, VTWO, VTWO, VTRE,
		 VONE, VTWO, VTWO, VTRE,
 		 VTWO, VTRE, VTRE, VFOR};

	/* count horizontal rows */
	y = 5; do{
		x = 6; do{
			chip = rack[x][y];
			shiftchip()
			/* wait until at lest 4 positions
			 * have been shifted in */
			if(x > 3) continue;
			evalpat()
		}while(x--);
	}while(y--);

	/* count vertical columns */
	x = 6; do{
		y = 5; do{
			chip = rack[x][y];
			shiftchip()
			/* wait until at lest 4 positions
			 * have been shifted in */
			if(y > 2) continue;
			evalpat()
		}while(y--);
	}while(x--);

	/* count rising diagonals */
	x = 3; do{
		y = 5; z = x + 5; do{
			if((z < 0) || (z > 6))
				shiftocc()
			else{
				chip = rack[z][y];
				shiftchip()
			}
			z--;
			/* wait until at lest 4 positions
			 * have been shifted in */
			if(y > 2) continue;
			evalpat()
		}while(y--);
	}while(x-- > -2);

	/* count falling diagonals */
	x = 8; do{
		y = 5; z = x - 5; do{
			if((z < 0) || (z > 6))
				shiftocc()
			else{
				chip = rack[z][y];
				shiftchip()
			}
			z++;
			/* wait until at lest 4 positions
			 * have been shifted in */
			if(y > 2) continue;
			evalpat()
		}while(y--);
	}while(x-- > 3);

	/* yellow and red advantage are treated equally.
	 * In this case the value of VNUL does not matter
	 * and should be set to 0 */
	return valyel - valred;
}


/* only works if yellow and red advantage are treated equally */
#define playervalue(player) ((player == CYEL)?value():(-value()))

int bmin(); /* forward declaration */


int cmax(min)
const int min;
{
	int w, cm = -MAXWINVALUE;
	register xtype c = 0;

	if(playervalue(opponent(player)) < VFOR) /* Opponent didn't yet win */
		do{
			if(spcclm[c]){
				inschip(c, player);
				if(actmove <= 1)
					w = playervalue(player);
				else{
					actmove--;
					w = bmin(cm); /* Recursion */
					if(w >= (VFOR - MAXSTRENGTH)) w--;
					actmove++;
				}
				if(w > cm)
					cm = w; /* new maximum */
				remchip(c);
			}
			if(cm >= min)
				c = 7; /* Alpha-Beta-Abschneidung */
		}while(++c < 7);
	return cm;
}


int bmin(min)
const int min;
{
	int w, bm = MAXWINVALUE;
	register xtype b = 0;

	if(playervalue(player) < VFOR) /* self didn't win */
		do{
			if(spcclm[b]){
				inschip(b, opponent(player));
				if(actmove <= 1)
					w = playervalue(player);
				else{
					actmove--;
					w = cmax(bm); /* Recursion */
					if(w <= -(VFOR - MAXSTRENGTH)) w++;
					actmove++;
				}
				if(w < bm)
					bm = w; /* new minimum */
				remchip(b);
			}
			if(bm <= min)
				b = 7; /* Alpha-Beta-Pruning */
		}while(++b < 7);
	return bm;
}


int minimax()
{
	int w, am = -MAXWINVALUE, minret = -1;
	xtype r = 6;
	register xtype a;

	if(spcrack > 40) /* Opening library */
		return 3;
	if(spcrack == 40){
		if((allmoves[0] == '4') && (allmoves[1] == '4'))
			return 2;
		if((allmoves[0] == '4') || (allmoves[1] == '4'))
			return 3;
	} /* End of opening library */
	do{
		a = sequence[r];
		if(spcclm[a]){
			inschip(a, player);
			actmove = strength - 1;
			w = bmin(am);
			if(w > am){
				am = w;
				minret = a;
			}
			remchip(a);
		}
	}while(r--);
	return minret;
}


void computermove()
{
	char    recommend;   /* recommended move */

	if(playervalue(CYEL) >= VFOR){
		winner |= CYEL;
		return;
	}
	if(playervalue(CRED) >= VFOR){
		winner |= CRED;
		return;
	}
	if(strength > MAXSTRENGTH)
		strength = MAXSTRENGTH;
	if(strength > spcrack)
		strength = spcrack;
	recommend = minimax();
	if((recommend < 0) && (recommend > 6))
		return;
	inschip(recommend, player);
	spcrack--;
	allmoves[42 - (spcrack + 1)] = recommend + '1';
	lastmoves[player - 1] = recommend;
	if(playervalue(player) >= VFOR)
		winner |= player;
	player = opponent(player);
}


/* clear the rack */
void clearrack()
{
	ytype y;
	xtype x;

	player = CYEL;
	winner = CEMT;
	y = 5; do{
		x = 6; do{
			rack[x][y] = CEMT;
		}while(x--);
	}while(y--);
}


#if defined(TESTMODE)
/* fill the rack in test mode
 * the query_string begins with '_' and contains
 * only '1' for yellow, '2' for red, '0' for empty
 * and '_' as a column separator */
int fillrack()
{
	const char *query_string;
	ytype y;
	xtype x;
	char i, c;

	if(!(query_string = getenv("QUERY_STRING"))) return 0;
	i = strlen(query_string);
	if((i % 8) != 0){
		error("invalid row/column-format");
		return -1;
	}
	y = 6 - (i / 8); do{
		x = 6; do{
			c = query_string[--i];
			if(!((c == '0') || (c == '1') || (c == '2'))){
				error("invalid chiptye");
				return -1;
			}
			rack[x][y] = c - '0';
			(spcclm[x])--;
			spcrack--;
		}while(x--);
		if(query_string[--i] != '_'){
			error("invalid row/column-format");
			return -1;
		}
		if(i <= 0) break;	
	}while(++y <= 5);
	return 1;
}


#else /* normal mode */
/* fill the rack in normal mode
 * the query_string contains only numbers between 1 ... 7
 * i. e. the columns where chips were inserted */
int fillrack()
{
	const char *query_string;
	char charclm;
	xtype numclm;
	int i = 0;


	if(!(query_string = getenv("QUERY_STRING"))) return 0;
	while(charclm = query_string[i]){
		allmoves[i] = charclm;
		numclm = charclm - '1';
		if(i > 41){
			error("No space left in rack");
			return -1;
		}
		if((numclm < 0) || (numclm > 6)){
			error("Invalid Column");
			return -1;
		}
		if(!spcclm[numclm]){
			error("No space left in column");
			return -1;
		}
		lastmoves[player - 1] = numclm;
		inschip(numclm, player);
		spcrack--;
		player = opponent(player);
		i++;
	}
	return 1;
}
#endif /* TESTMODE */


/* writes a line of the rack to the browsers window */
void writerackline(y)
const ytype y;
{
	register xtype x;

	printf("<TR>\n");
	x = 0; do{
		printf("<TD>");
		switch(rack[x][y]){
		case CEMT:
			printf("<IMG SRC=\"/W4Row/img/Blu.gif\" WIDTH=32 HEIGHT=32, ALT=\".\">");
			break;
		case CYEL:
			printf("<IMG SRC=\"/W4Row/img/Yel.gif\" WIDTH=32 HEIGHT=32, ALT=\"X\">");
			break;
		case CRED:
			printf("<IMG SRC=\"/W4Row/img/Red.gif\" WIDTH=32 HEIGHT=32, ALT=\"O\">");
			break;
		}
		printf("</TD>\n");
	}while(++x <= 6);
	printf("</TR>\n");
}


/* writes an HTML-table header above the rack */
void writetableheader()
{
	register xtype x;
	char charclm;

	printf("<TR>\n");
	x = 0; do{
		charclm = '1' + x;
		if(spcclm[x] && !winner){
			printf("<TD><A HREF=\"/cgi-bin/W4Row.exe?%s%c\">", allmoves, charclm);
			printf("<IMG SRC=\"/W4Row/img/Button.gif\" ALT= \"%c\" WIDTH=28 HEIGHT=12></A></TD>\n", charclm);
		}else
			printf("<TD>%c</TD>\n", charclm);
	}while(++x <= 6);
	/* if(((42-spcrack) & 1) == 1)
		if(lastmove[0] != (allmoves[42-spcrack] - '0'))
			error("program error");
	if(((42-spcrack) & 1) == 0)
		if(lastmove[1] != (allmoves[42-spcrack] - '0'))
			error("program error"); */
	printf("</TR>\n");
	printf("<TR>\n");
	x = 0; do{
		printf("<TD><IMG SRC=\"/W4Row/img/");
		if(lastmoves[0] == x)
			printf("Yelsm");
		else
			printf("Blusm");
		printf(".gif\" WIDTH=16 HEIGHT=16>");
		printf("<IMG SRC=\"/W4Row/img/");
		if(lastmoves[1] == x)
			printf("Redsm");
		else
			printf("Blusm");
		printf(".gif\" WIDTH=16 HEIGHT=16></TD>\n");
	}while(++x <= 6);
	printf("</TR>\n");
}


/* writes the whole rack to the browsers window */
void writefeld()
{
	register ytype y;

	printf("Content-type: text/html\n\n");
	printf("<HTML>\n<HEAD>\n<TITLE>W4Row</TITLE>\n</HEAD>\n");
	printf("<BODY BGCOLOR=\"#C0C0C0\">\n");
	if(winner & CRED)
		printf("<H1>Red won the Game</H1>\n");
	if(winner & CYEL)
		printf("<H1>Yellow won the Game</H1>\n");
	if(!winner){
		printf("<H1>W4Row ");
		if(player == CRED)
			printf("(Red's move)</H1>\n");
		if(player == CYEL)
			printf("(Yellow's move)</H1>\n");
	}
	printf("<CENTER>\n<TABLE BORDER=2>\n");
	writetableheader();
	y = 0;	do{
		writerackline(y);
	}while(++y <= 5);
	printf("</TABLE>\n</CENTER>\n");
#if defined(TESTMODE)
	printf("Value = %d\n", value());
#endif
	printf("<H2>Computer shall <A HREF=\"/cgi-bin/W4Row.exe?%s\">move</A> / ", allmoves);
	printf("<A HREF=\"/cgi-bin/W4Row.exe\">New</A> Game / ");
	printf("<A HREF=\"/W4Row/index.html\">Home</A></H2>\n");
	printf("<ADDRESS>Version 2.0, Copyright 1997 %s</ADDRESS>\n</BODY>\n</HTML>\n", MYADDRESS);
}


int main(argc, argv)
int argc;
char *argv[];
{
	clearrack();
	if(fillrack() < 0)
		return -1;
#if !defined(TESTMODE)
	computermove();
#endif
	writefeld();
	return 0;
}
